#!/usr/bin/env python
# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
stack_hasher will take input text, parse out the stack trace and hash based on
the stack. When using to bucket crashes it is helpful to make two hashes, one
of the entire stack and the second that is less sensitive using the first few
entries on the top of the stack with the offsets removed. This returns a unique
crash id (1st hash) and a bug id (2nd hash). This is not perfect but works very
well in most cases.
"""
from hashlib import sha1
from logging import basicConfig, getLogger, INFO, DEBUG
from os.path import basename
from re import compile as re_compile, match as re_match

__all__ = ("Stack", "StackFrame")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger("stack_hasher")

MAJOR_DEPTH = 5
MAJOR_DEPTH_RUST = 10


class StackFrame(object):
    MODE_GDB = 0
    MODE_MINIDUMP = 1
    MODE_RR = 2
    MODE_RUST = 3
    MODE_SANITIZER = 4
    MODE_TSAN = 5
    MODE_VALGRIND = 6

    _re_func_name = re_compile(r"(?P<func>.+?)[\(|\s|\<]{1}")
    # regexs for supported stack trace lines
    _re_gdb = re_compile(r"^#(?P<num>\d+)\s+(?P<off>0x[0-9a-f]+\sin\s)*(?P<line>.+)")
    _re_rr = re_compile(r"rr\((?P<loc>.+)\+(?P<off>0x[0-9a-f]+)\)\[0x[0-9a-f]+\]")
    _re_rust_frame = re_compile(r"^\s+(?P<num>\d+):\s+0x[0-9a-f]+\s+\-\s+(?P<line>.+)")
    _re_sanitizer = re_compile(r"^\s*#(?P<num>\d+)\s0x[0-9a-f]+(?P<in>\sin)?\s+(?P<line>.+)")
    _re_tsan = re_compile(r"^\s*#(?P<num>\d+)\s(?P<line>.+)\s\(((?P<mod>.+)\+)?(?P<off>0x[0-9a-f]+)\)")
    _re_valgrind = re_compile(r"^==\d+==\s+(at|by)\s+0x[0-9A-F]+\:\s+(?P<func>.+?)\s+\((?P<line>.+)\)")
    # TODO: add additional debugger support?
    #_re_rust_file = re_compile(r"^\s+at\s+(?P<line>.+)")
    #_re_windbg = re_compile(r"^(\(Inline\)|[a-f0-9]+)\s([a-f0-9]+|-+)\s+(?P<line>.+)\+(?P<off>0x[a-f0-9]+)")

    __slots__ = ("function", "location", "mode", "offset", "stack_line")

    def __init__(self, function=None, location=None, mode=None, offset=None, stack_line=None):
        self.function = function
        self.location = location
        self.mode = mode
        self.offset = offset
        self.stack_line = stack_line

    def __str__(self):
        out = []
        if self.stack_line is not None:
            out.append("%02d" % int(self.stack_line))
        if self.function is not None:
            out.append("function: %r" % self.function)
        if self.location is not None:
            out.append("location: %r" % self.location)
        if self.offset is not None:
            out.append("offset: %r" % self.offset)
        return " - ".join(out)

    @classmethod
    def from_line(cls, input_line, parse_mode=None):
        assert "\n" not in input_line, "Input contains unexpected new line(s)"
        sframe = None
        if parse_mode is None or parse_mode == cls.MODE_SANITIZER:
            sframe = cls._parse_sanitizer(input_line)
        if not sframe and parse_mode is None or parse_mode == cls.MODE_GDB:
            sframe = cls._parse_gdb(input_line)
        if not sframe and parse_mode is None or parse_mode == cls.MODE_MINIDUMP:
            sframe = cls._parse_minidump(input_line)
        if not sframe and parse_mode is None or parse_mode == cls.MODE_RR:
            sframe = cls._parse_rr(input_line)
        if not sframe and parse_mode is None or parse_mode == cls.MODE_RUST:
            sframe = cls._parse_rust(input_line)
        if not sframe and parse_mode is None or parse_mode == cls.MODE_TSAN:
            sframe = cls._parse_tsan(input_line)
        if not sframe and parse_mode is None or parse_mode == cls.MODE_VALGRIND:
            sframe = cls._parse_valgrind(input_line)
        return sframe

    @classmethod
    def _parse_gdb(cls, input_line):
        if "#" not in input_line:
            return None
        m = cls._re_gdb.match(input_line)
        if m is None:
            return None
        input_line = m.group("line").strip()
        if not input_line:
            return None
        sframe = cls(mode=cls.MODE_GDB, stack_line=m.group("num"))
        #sframe.offset = m.group("off")  # ignore binary offset for now
        # find function/method name
        m = cls._re_func_name.match(input_line)
        if m is not None:
            sframe.function = m.group("func")
        # find file name and line number
        if ") at " in input_line:
            input_line = input_line.split(") at ")[-1]
            try:
                input_line, sframe.offset = input_line.split(":")
            except ValueError:
                pass
            sframe.location = basename(input_line).split()[0]
        return sframe

    @classmethod
    def _parse_minidump(cls, input_line):
        try:
            tid, stack_line, lib_name, func_name, file_name, line_no, offset = input_line.split("|")
            if int(tid) < 0 or int(stack_line) < 0:
                return None
        except ValueError:
            return None
        sframe = cls(mode=cls.MODE_MINIDUMP, stack_line=stack_line)
        if func_name:
            sframe.function = func_name.strip()
        if file_name:
            if file_name.count(":") > 1:
                # contains hg repo info
                sframe.location = basename(file_name.split(":")[-2])
            else:
                sframe.location = file_name
        elif lib_name:
            sframe.location = lib_name.strip()
        if line_no:
            sframe.offset = line_no.strip()
        elif offset:
            sframe.offset = offset.strip()
        return sframe

    @classmethod
    def _parse_rr(cls, input_line):
        if "rr(" not in input_line:
            return None
        m = cls._re_rr.match(input_line)
        if m is None:
            return None
        return cls(location=m.group("loc"), mode=cls.MODE_RR, offset=m.group("off"))

    @classmethod
    def _parse_rust(cls, input_line):
        m = cls._re_rust_frame.match(input_line)
        if m is None:
            return None
        sframe = cls(mode=cls.MODE_RUST, stack_line=m.group("num"))
        sframe.function = m.group("line").strip().rsplit("::h", 1)[0]
        # Don't bother with the file offset stuff atm
        #m = cls._re_rust_file.match(input_line) if frame is None else None
        #if m is not None:
        #    frame = {"function":None, "mode":cls.MODE_RUST, "offset":None, "stack_line":None}
        #    input_line = m.group("line").strip()
        #    if ":" in input_line:
        #        frame["location"], frame["offset"] = input_line.rsplit(":", 1)
        #    else:
        #        frame["location"] = input_line
        return sframe

    @classmethod
    def _parse_sanitizer(cls, input_line):
        if "#" not in input_line:
            return None
        m = cls._re_sanitizer.match(input_line)
        if m is None:
            return None
        sframe = cls(mode=cls.MODE_SANITIZER, stack_line=m.group("num"))
        input_line = m.group("line")
        # check if line is symbolized
        if m.group("in"):
            # find function/method name
            m = cls._re_func_name.match(input_line)
            if m is not None:
                sframe.function = m.group("func")
        if input_line.startswith("("):
            input_line = input_line.strip("()")
        # find location (file name or module) and offset (line # or offset)
        offset = re_match(r"(.+?)(\:([0-9a-f]+)|\+(0x[0-9a-f]+)).*", input_line)
        if offset:
            sframe.location = basename(offset.group(1))
            sframe.offset = offset.group(3) or offset.group(4)
        else:
            sframe.location = input_line
        return sframe

    @classmethod
    def _parse_tsan(cls, input_line):
        if "#" not in input_line:
            return None
        m = cls._re_tsan.match(input_line)
        if m is None:
            return None
        sframe = cls(mode=cls.MODE_TSAN, stack_line=m.group("num"))
        input_line = m.group("line")
        location = basename(input_line)
        # try to parse file name and line number
        if location:
            location = location.split()[-1].split(":")
            if location and location[0] != "<null>":
                sframe.location = location.pop(0)
                if location and location[0] != "<null>":
                    sframe.offset = location.pop(0)
        # use module name if file name cannot be found
        if not sframe.location:
            sframe.location = m.group("mod")
        # use module offset if line number cannot be found
        if not sframe.offset:
            sframe.offset = m.group("off")
        m = cls._re_func_name.match(input_line)
        if m is not None:
            function = m.group("func")
            if function and function != "<null>":
                sframe.function = function
        return sframe

    @classmethod
    def _parse_valgrind(cls, input_line):
        if "== " not in input_line:
            return None
        m = cls._re_valgrind.match(input_line)
        if m is None:
            return None
        input_line = m.group("line")
        if input_line is None:  # pragma: no cover
            # this should not happen
            LOG.warning("failure in _parse_valgrind()")
            return None
        sframe = cls(function=m.group("func"), mode=cls.MODE_VALGRIND)
        try:
            location, sframe.offset = input_line.split(":")
            sframe.location = location.strip()
        except ValueError:
            # trim anything from the beginning we might have missed
            location = input_line.rsplit("(")[-1]
            if location.startswith("in "):
                location = input_line[3:]
            sframe.location = basename(location)
        if not sframe.location:
            return None
        return sframe


class Stack(object):
    __slots__ = ("frames", "_major", "_major_depth", "_minor")

    def __init__(self, frames=None, major_depth=MAJOR_DEPTH):
        assert frames is None or isinstance(frames, list)
        self.frames = list() if frames is None else frames
        self._major_depth = major_depth
        self._major = None
        self._minor = None

    def __str__(self):
        return "\n".join(str(frame) for frame in self.frames)

    def _calculate_hash(self, major=False):
        if not self.frames or (major and self._major_depth < 1):
            return None
        h = sha1()
        current_depth = 0
        for frame in self.frames:
            current_depth += 1
            if major and current_depth > self._major_depth:
                break
            if frame.location is not None:
                h.update(frame.location.encode("utf-8", errors="ignore"))
            if frame.function is not None:
                h.update(frame.function.encode("utf-8", errors="ignore"))
            if major and current_depth > 1:
                # only add the offset from the top frame when calculating
                # the major hash and skip the rest
                continue
            if frame.offset is not None:
                h.update(frame.offset.encode("utf-8", errors="ignore"))
        return h.hexdigest()

    def from_file(self, file_name):  # pragma: no cover
        raise NotImplementedError()  # TODO

    @classmethod
    def from_text(cls, input_text, major_depth=MAJOR_DEPTH, parse_mode=None):
        """
        parse a stack trace from text.
        input_txt is the data to parse the trace from.
        """

        frames = list()
        prev_line = None
        for line in reversed(input_text.split("\n")):
            if not line:
                # skip empty lines
                continue
            try:
                frame = StackFrame.from_line(line, parse_mode=parse_mode)
            except Exception:  # pragma: no cover
                LOG.error("Error calling from_line() with: %r", line)
                raise
            if frame is None:
                continue

            # avoid issues with mixed stack types
            if parse_mode is None:
                parse_mode = frame.mode
            elif parse_mode != frame.mode:
                # don't mix parse modes!
                continue

            if frame.stack_line is not None:
                stack_line = int(frame.stack_line)
                # check if we've found a different stack in the data
                if prev_line is not None and prev_line <= stack_line:
                    break
                frames.insert(0, frame)
                if stack_line < 1:
                    break
                prev_line = stack_line
            else:
                frames.insert(0, frame)

        # sanity check
        if frames and prev_line is not None:
            # assuming the first frame is 0
            if int(frames[0].stack_line) != 0:
                LOG.warning("First stack line %s not 0", frames[0].stack_line)
            if int(frames[-1].stack_line) != len(frames) - 1:
                LOG.warning("Last stack line %s not %d (frames-1)", frames[0].stack_line, len(frames) - 1)

        if frames and frames[0].mode == StackFrame.MODE_RUST and major_depth < MAJOR_DEPTH_RUST:
            major_depth = MAJOR_DEPTH_RUST

        return cls(frames=frames, major_depth=major_depth)

    @property
    def major(self):
        if self._major is None:
            self._major = self._calculate_hash(major=True)
        return self._major

    @property
    def minor(self):
        if self._minor is None:
            self._minor = self._calculate_hash()
        return self._minor


if __name__ == "__main__":
    from argparse import ArgumentParser
    from os import getenv  # pylint: disable=ungrouped-imports

    parser = ArgumentParser()
    parser.add_argument("input", help="")
    args = parser.parse_args()

    # set output verbosity
    if getenv("DEBUG"):
        log_level = DEBUG
        log_fmt = "[%(levelname).1s] %(message)s"
    else:
        log_level = INFO
        log_fmt = "%(message)s"
    basicConfig(format=log_fmt, datefmt="%Y-%m-%d %H:%M:%S", level=log_level)

    with open(args.input, "rb") as fp:
        stack = Stack.from_text(fp.read().decode("utf-8", errors="ignore"))

    for frame in stack.frames:
        LOG.info(frame)
    LOG.info("Minor: %s", stack.minor)
    LOG.info("Major: %s", stack.major)
    LOG.info("Frames: %d", len(stack.frames))



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""test Grizzly main"""

from pytest import raises

from sapphire import Sapphire
from .common import Adapter
from .main import main
from .session import Session
from .target import TargetLaunchError


class FakeArgs(object):
    def __init__(self, working_path):
        self.binary = None
        self.input = None
        self.adapter = None
        self.cache = 0
        self.coverage = False
        self.extension = None
        self.fuzzmanager = False
        self.ignore = list()
        self.launch_timeout = 300
        self.log_level = 10  # 10 = DEBUG, 20 = INFO
        self.log_limit = 0
        self.memory = 0
        self.platform = "test"
        self.prefs = None
        self.rr = False
        self.relaunch = 1000
        self.s3_fuzzmanager = False
        self.timeout = 60
        self.tool = None
        self.valgrind = False
        self.verbose = False
        self.working_path = working_path
        self.xvfb = False

# TODO: these could use call_count checks

def test_main_01(tmp_path, mocker):
    """test main()"""
    fake_adapter = mocker.Mock(spec=Adapter)
    fake_adapter.NAME = "fake"
    fake_adapter.RELAUNCH = 1
    fake_adapter.TEST_DURATION = 10
    mocker.patch("grizzly.main.get_adapter", return_value=lambda: fake_adapter)
    targets = mocker.patch("grizzly.target.TARGETS")
    targets.return_value = "fake-target"
    fake_session = mocker.patch("grizzly.main.Session", autospec=True)
    fake_session.return_value.server = mocker.Mock(spec=Sapphire)
    fake_session.EXIT_SUCCESS = Session.EXIT_SUCCESS
    args = FakeArgs(str(tmp_path))
    args.adapter = "fake"
    args.coverage = True
    args.input = "fake"
    args.ignore = ["fake", "fake"]
    args.rr = True
    args.valgrind = True
    args.xvfb = True
    with raises(RuntimeError, match="Coverage must be run with --relaunch > 1"):
        main(args)
    fake_adapter.RELAUNCH = 0
    assert main(args) == Session.EXIT_SUCCESS
    fake_reporter = mocker.patch("grizzly.main.FuzzManagerReporter", autospec=True)
    fake_reporter.sanity_check.return_value = True
    args.input = None
    args.log_level = None
    args.fuzzmanager = True
    assert main(args) == Session.EXIT_SUCCESS
    fake_reporter = mocker.patch("grizzly.main.S3FuzzManagerReporter", autospec=True)
    fake_reporter.sanity_check.return_value = True
    args.fuzzmanager = False
    args.s3_fuzzmanager = True
    assert main(args) == Session.EXIT_SUCCESS

def test_main_02(tmp_path, mocker):
    """test main()"""
    fake_adapter = mocker.Mock(spec=Adapter)
    mocker.patch("grizzly.main.get_adapter", return_value=lambda: fake_adapter)
    fake_session = mocker.patch("grizzly.main.Session", autospec=True)
    fake_session.EXIT_SUCCESS = Session.EXIT_SUCCESS
    args = FakeArgs(str(tmp_path))
    args.adapter = "fake"
    fake_adapter.TEST_DURATION = args.timeout + 10
    with raises(RuntimeError, match=r"Test duration \([0-9]+s\) should be less than browser timeout \([0-9]+s\)"):
        main(args)

def test_main_03(tmp_path, mocker):
    """test main() exit codes"""
    fake_adapter = mocker.Mock(spec=Adapter)
    fake_adapter.TEST_DURATION = 10
    fake_adapter.RELAUNCH = 0
    mocker.patch("grizzly.main.get_adapter", return_value=lambda: fake_adapter)
    targets = mocker.patch("grizzly.target.TARGETS")
    targets.return_value = "fake-target"
    fake_session = mocker.patch("grizzly.main.Session", autospec=True)
    fake_session.EXIT_SUCCESS = Session.EXIT_SUCCESS
    fake_session.EXIT_ABORT = Session.EXIT_ABORT
    fake_session.EXIT_LAUNCH_FAILURE = Session.EXIT_LAUNCH_FAILURE
    fake_session.return_value.server = mocker.Mock(spec=Sapphire)
    args = FakeArgs(str(tmp_path))
    args.adapter = "fake"
    args.input = "fake"
    fake_session.return_value.run.side_effect = KeyboardInterrupt
    assert main(args) == Session.EXIT_ABORT
    fake_session.return_value.run.side_effect = TargetLaunchError("test")
    assert main(args) == Session.EXIT_LAUNCH_FAILURE



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
from __future__ import unicode_literals
import os
import time
import pytest
import sapphire
from ..reduce.reduce import ReductionJob
from ..target.target import Target, TargetLaunchError, TargetLaunchTimeout
from .test_common import FakeTarget, FakeReduceStatus, create_target_binary


class FakeServer(object):
    _last_timeout = None

    def __init__(self, **kwds):
        FakeServer._last_timeout = kwds.get("timeout")

    def close(self):
        pass

    @property
    def port(self):  # pylint: disable=no-self-use
        return 8000

    def add_dynamic_response(self, *args, **kwds):
        pass

    def set_redirect(self, *args, **kwds):
        pass

    def serve_testcase(self, testcase, **kwds):  # pylint: disable=no-self-use,unused-argument
        testcase.duration = 0.1
        return sapphire.SERVED_ALL, ["test.html"]

    @property
    def timeout(self):
        return FakeServer._last_timeout

    @timeout.setter
    def timeout(self, value):  # pylint: disable=no-self-use
        assert value is None or value >= 0
        FakeServer._last_timeout = value


@pytest.fixture
def fake_sapphire(monkeypatch):
    monkeypatch.setattr(sapphire, "Sapphire", FakeServer)
    FakeServer._last_timeout = None

pytestmark = pytest.mark.usefixtures("fake_sapphire")


def test_interesting(tmp_path):
    "simple case where the test is interesting"
    obj = ReductionJob([], FakeTarget(), 30, False, False, 0, 1, 1, 0, 0, FakeReduceStatus())
    create_target_binary(obj.target, tmp_path)
    (tmp_path / "test.html").touch()
    obj.reduce_file = str(tmp_path / "test.html")
    obj.lithium_init()
    (tmp_path / "lithium").mkdir()
    assert obj.lithium_interesting(str(tmp_path / "lithium"))
    assert "grz_close_browser" in obj._server_map.dynamic
    assert "grz_harness" in obj._server_map.dynamic
    assert "grz_current_test" in obj._server_map.redirect
    assert "grz_next_test" in obj._server_map.redirect
    assert obj.target._calls["close"] == 1
    obj.lithium_cleanup()
    assert obj.target._calls["cleanup"] == 0
    assert obj.target._calls["detect_failure"] == 1


def test_not_interesting(tmp_path):
    "simple case where the test is not interesting"

    class MyTarget(FakeTarget):

        def detect_failure(self, *args, **kwds):
            FakeTarget.detect_failure(self, *args, **kwds)
            return Target.RESULT_NONE

    obj = ReductionJob([], MyTarget(), 30, False, False, 0, 1, 1, 0, 0, FakeReduceStatus())
    create_target_binary(obj.target, tmp_path)
    obj.target._is_healthy = True
    (tmp_path / "test.html").touch()
    obj.reduce_file = str(tmp_path / "test.html")
    obj.lithium_init()
    (tmp_path / "lithium").mkdir()
    assert not obj.lithium_interesting(str(tmp_path / "lithium"))
    assert obj.server is not None
    assert obj.target._calls["close"] == 0
    obj.lithium_cleanup()
    assert obj.target._calls["cleanup"] == 0
    assert obj.target._calls["detect_failure"] == 1


def test_ignored(tmp_path):
    "if target says ignored, it's not interesting"

    class MyTarget(FakeTarget):

        def detect_failure(self, ignored, *args, **kwds):
            FakeTarget.detect_failure(self, *args, **kwds)
            assert ignored == ["foo", "bar"]
            return Target.RESULT_IGNORED

    obj = ReductionJob(["foo", "bar"], MyTarget(), 30, False, False, 0, 1, 1, 0, 0, FakeReduceStatus())
    create_target_binary(obj.target, tmp_path)
    (tmp_path / "test.html").touch()
    obj.reduce_file = str(tmp_path / "test.html")
    obj.lithium_init()
    (tmp_path / "lithium").mkdir()
    assert not obj.lithium_interesting(str(tmp_path / "lithium"))
    assert obj.server is not None
    assert obj.target._calls["close"] == 1
    obj.lithium_cleanup()
    assert obj.target._calls["cleanup"] == 0
    assert obj.target._calls["detect_failure"] == 1


def test_target_relaunch_error(tmp_path):
    "target should be launched only once on TargetLaunchError"

    class MyTarget(FakeTarget):

        def launch(self, *args, **kwds):
            FakeTarget.launch(self, *args, **kwds)
            raise TargetLaunchError()

    obj = ReductionJob([], MyTarget(), 30, False, False, 0, 1, 1, 0, 0, FakeReduceStatus())
    create_target_binary(obj.target, tmp_path)
    prefix = tmp_path / "lithium"
    prefix.mkdir()
    (tmp_path / "test.html").touch()
    obj.reduce_file = str(tmp_path / "test.html")
    obj.lithium_init()
    with pytest.raises(TargetLaunchError):
        obj.lithium_interesting(str(prefix))
    assert obj.server is not None
    assert obj.target._calls["launch"] == 1
    obj.lithium_cleanup()
    assert obj.target._calls["cleanup"] == 0


def test_target_relaunch_timeout(mocker, tmp_path):
    "target should be launched more than once on TargetLaunchTimeout"

    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    class MyTarget(FakeTarget):

        def launch(self, *args, **kwds):
            FakeTarget.launch(self, *args, **kwds)
            raise TargetLaunchTimeout()

    obj = ReductionJob([], MyTarget(), 30, False, False, 0, 1, 1, 0, 0, FakeReduceStatus())
    create_target_binary(obj.target, tmp_path)
    prefix = tmp_path / "lithium"
    prefix.mkdir()
    (tmp_path / "test.html").touch()
    obj.reduce_file = str(tmp_path / "test.html")
    obj.lithium_init()
    with pytest.raises(TargetLaunchTimeout):
        obj.lithium_interesting(str(prefix))
    assert obj.server is not None
    assert obj.target._calls["launch"] > 1
    obj.lithium_cleanup()
    assert obj.target._calls["cleanup"] == 0


def test_no_harness(tmp_path):
    "simple case where harness is not used"
    obj = ReductionJob([], FakeTarget(), 30, True, False, 0, 1, 1, 0, 0, FakeReduceStatus())
    create_target_binary(obj.target, tmp_path)
    prefix = tmp_path / "lithium"
    prefix.mkdir()
    reduce_file = tmp_path / "test.html"
    reduce_file.touch()
    obj.reduce_file = str(reduce_file)
    obj.lithium_init()
    assert obj.lithium_interesting(str(prefix))
    assert "grz_close_browser" not in obj._server_map.dynamic
    assert "grz_harness" not in obj._server_map.dynamic
    assert "grz_current_test" in obj._server_map.redirect
    assert "grz_next_test" not in obj._server_map.redirect
    assert obj.target._calls["close"] == 1
    obj.lithium_cleanup()
    assert obj.target._calls["cleanup"] == 0
    assert obj.target._calls["detect_failure"] == 1


def test_skip(tmp_path):
    "skip should assume interesting for the first n calls, without launching the target"
    obj = ReductionJob([], FakeTarget(), 30, False, False, 7, 1, 1, 0, 0, FakeReduceStatus(),
                       testcase_cache=False)
    create_target_binary(obj.target, tmp_path)
    reduce_file = tmp_path / "test.html"
    reduce_file.touch()
    obj.reduce_file = str(reduce_file)
    obj.lithium_init()
    # first call is real, regardless of skip, since that is the initial repro
    prefix = tmp_path / "lithium0"
    prefix.mkdir()
    assert obj.lithium_interesting(str(prefix))
    assert obj.target._calls["launch"] == 1
    assert obj.target._calls["close"] == 1
    for _ in range(7):
        assert not obj.lithium_interesting(None)
    assert obj.target._calls["launch"] == 1
    prefix = tmp_path / "lithium1"
    prefix.mkdir()
    assert obj.lithium_interesting(str(prefix))
    assert obj.target._calls["launch"] == 2
    assert obj.target._calls["close"] == 2
    assert obj.server is not None
    obj.lithium_cleanup()
    assert obj.target._calls["cleanup"] == 0


def test_any_crash_false(tmp_path):
    "subsequent crashes must match the original if any_crash is False"
    stderr = "Assertion failure: bad thing happened, at test.c:123"

    class MyTarget(FakeTarget):

        def save_logs(self, dest, **kwds):
            FakeTarget.save_logs(self, dest, **kwds)
            with open(os.path.join(dest, "log_stderr.txt"), "w") as log_fp:
                log_fp.write(stderr)

    obj = ReductionJob([], MyTarget(), 30, False, False, 0, 1, 1, 0, 0, FakeReduceStatus(),
                       testcase_cache=False)
    create_target_binary(obj.target, tmp_path)
    reduce_file = tmp_path / "test.html"
    reduce_file.touch()
    obj.reduce_file = str(reduce_file)
    obj.lithium_init()
    # first call is real, regardless of skip, since that is the initial repro
    prefix = tmp_path / "lithium0"
    prefix.mkdir()
    assert obj.lithium_interesting(str(prefix))
    assert obj.target._calls["launch"] == 1
    assert obj.target._calls["close"] == 1
    prefix = tmp_path / "lithium1"
    prefix.mkdir()
    stderr = "Assertion failure: some other thing happened, at test.c:456"
    assert not obj.lithium_interesting(str(prefix))  # doesn't match original sig
    assert obj.target._calls["launch"] == 2
    assert obj.target._calls["close"] == 2
    assert obj.server is not None
    obj.lithium_cleanup()
    assert obj.target._calls["cleanup"] == 0


def test_any_crash_true(tmp_path):
    "subsequent crashes need not match the original if any_crash is True"
    stderr = "Assertion failure: bad thing happened, at test.c:123"

    class MyTarget(FakeTarget):

        def save_logs(self, dest, **kwds):
            FakeTarget.save_logs(self, dest, **kwds)
            with open(os.path.join(dest, "log_stderr.txt"), "w") as log_fp:
                log_fp.write(stderr)

    obj = ReductionJob([], MyTarget(), 30, False, True, 0, 1, 1, 0, 0, FakeReduceStatus(),
                       testcase_cache=False)
    create_target_binary(obj.target, tmp_path)
    reduce_file = tmp_path / "test.html"
    reduce_file.touch()
    obj.reduce_file = str(reduce_file)
    obj.lithium_init()
    # first call is real, regardless of skip, since that is the initial repro
    prefix = tmp_path / "lithium0"
    prefix.mkdir()
    assert obj.lithium_interesting(str(prefix))
    assert obj.target._calls["launch"] == 1
    assert obj.target._calls["close"] == 1
    prefix = tmp_path / "lithium1"
    prefix.mkdir()
    stderr = "Assertion failure: some other thing happened, at test.c:456"
    assert obj.lithium_interesting(str(prefix))
    assert obj.target._calls["launch"] == 2
    assert obj.target._calls["close"] == 2
    assert obj.server is not None
    obj.lithium_cleanup()
    assert obj.target._calls["cleanup"] == 0


def test_min_crashes_repro(tmp_path):
    "min_crashes will force n iterations if crash repros"
    obj = ReductionJob([], FakeTarget(), 30, False, False, 0, 4, 1, 0, 0, FakeReduceStatus(),
                       testcase_cache=False)
    create_target_binary(obj.target, tmp_path)
    reduce_file = tmp_path / "test.html"
    reduce_file.touch()
    obj.reduce_file = str(reduce_file)
    obj.lithium_init()
    prefix = tmp_path / "lithium"
    prefix.mkdir()
    assert obj.lithium_interesting(str(prefix))
    assert obj.target._calls["launch"] == 4
    assert obj.target._calls["close"] == 4
    assert obj.server is not None
    assert obj.target._calls["detect_failure"] == 4
    obj.lithium_cleanup()
    assert obj.target._calls["cleanup"] == 0


def test_min_crashes_norepro(tmp_path):
    "min_crashes will return uninteresting if crash repros n-1 times"

    class MyTarget(FakeTarget):
        repros = 0

        def detect_failure(self, *args, **kwds):
            FakeTarget.detect_failure(self, *args, **kwds)
            if self.repros == 3:
                return Target.RESULT_NONE
            self.repros += 1
            return Target.RESULT_FAILURE

    obj = ReductionJob([], MyTarget(), 30, False, False, 0, 4, 1, 0, 0, FakeReduceStatus(),
                       testcase_cache=False)
    create_target_binary(obj.target, tmp_path)
    reduce_file = tmp_path / "test.html"
    reduce_file.touch()
    obj.reduce_file = str(reduce_file)
    obj.lithium_init()
    prefix = tmp_path / "lithium"
    prefix.mkdir()
    assert not obj.lithium_interesting(str(prefix))
    assert obj.target._calls["launch"] == 4
    assert obj.target._calls["close"] == 3
    assert obj.server is not None
    assert obj.target._calls["detect_failure"] == 4
    obj.lithium_cleanup()
    assert obj.target._calls["cleanup"] == 0


def test_repeat_repro(tmp_path):
    "repeat will stop at min_crashes if crash repros"
    obj = ReductionJob([], FakeTarget(), 30, False, False, 0, 4, 17, 0, 0, FakeReduceStatus(),
                       testcase_cache=False)
    create_target_binary(obj.target, tmp_path)
    reduce_file = tmp_path / "test.html"
    reduce_file.touch()
    obj.reduce_file = str(reduce_file)
    obj.lithium_init()
    prefix = tmp_path / "lithium"
    prefix.mkdir()
    assert obj.lithium_interesting(str(prefix))
    assert obj.server is not None
    assert obj.target._calls["detect_failure"] == 4
    obj.lithium_cleanup()
    assert obj.target._calls["cleanup"] == 0


def test_repeat_norepro(tmp_path):
    "repeat will force repeat - min_crashes + 1 iterations if crash doesn't repro"

    class MyTarget(FakeTarget):

        def detect_failure(self, *args, **kwds):
            FakeTarget.detect_failure(self, *args, **kwds)
            return Target.RESULT_NONE

    obj = ReductionJob([], MyTarget(), 30, False, False, 0, 4, 17, 0, 0, FakeReduceStatus(),
                       testcase_cache=False)
    create_target_binary(obj.target, tmp_path)
    reduce_file = tmp_path / "test.html"
    reduce_file.touch()
    obj.reduce_file = str(reduce_file)
    obj.lithium_init()
    prefix = tmp_path / "lithium"
    prefix.mkdir()
    assert not obj.lithium_interesting(str(prefix))
    assert obj.server is not None
    assert obj.target._calls["detect_failure"] == 17 - 4 + 1
    obj.lithium_cleanup()
    assert obj.target._calls["cleanup"] == 0


def test_cache(tmp_path):
    "cache will be hit if same file is given twice"
    obj = ReductionJob([], FakeTarget(), 30, False, False, 0, 1, 1, 0, 0, FakeReduceStatus())
    create_target_binary(obj.target, tmp_path)
    (tmp_path / "test.html").touch()
    obj.reduce_file = str(tmp_path / "test.html")
    obj.lithium_init()
    (tmp_path / "lithium0").mkdir()
    assert obj.lithium_interesting(tmp_path / "lithium0")
    (tmp_path / "lithium1").mkdir()
    assert obj.lithium_interesting(tmp_path / "lithium1")
    assert obj.target._calls["detect_failure"] == 1
    obj.lithium_cleanup()


def test_no_cache(tmp_path):
    "testcase_cache=False will disable cache"
    obj = ReductionJob([], FakeTarget(), 30, False, False, 0, 1, 1, 0, 0, FakeReduceStatus(),
                       testcase_cache=False)
    create_target_binary(obj.target, tmp_path)
    (tmp_path / "test.html").touch()
    obj.reduce_file = str(tmp_path / "test.html")
    obj.lithium_init()
    (tmp_path / "lithium0").mkdir()
    assert obj.lithium_interesting(tmp_path / "lithium0")
    (tmp_path / "lithium1").mkdir()
    assert obj.lithium_interesting(tmp_path / "lithium1")
    assert obj.target._calls["detect_failure"] == 2
    obj.lithium_cleanup()


def test_timeout_update(monkeypatch, tmp_path):
    "timeout will be updated based on time to crash"
    monkeypatch.setattr(time, "time", lambda: 0)  # run time will be calculated as 0
    failure_result = Target.RESULT_NONE

    class MyTarget(FakeTarget):

        def detect_failure(self, *args, **kwds):
            FakeTarget.detect_failure(self, *args, **kwds)
            return failure_result

    obj = ReductionJob([], MyTarget(), 30, False, False, 0, 1, 1, 0, 30, FakeReduceStatus(),
                       testcase_cache=False)
    create_target_binary(obj.target, tmp_path)
    (tmp_path / "test.html").touch()
    obj.reduce_file = str(tmp_path / "test.html")
    obj.lithium_init()
    assert not obj._fixed_timeout
    assert obj._idle_timeout == 30
    (tmp_path / "lithium0").mkdir()
    assert not obj.lithium_interesting(tmp_path / "lithium0")
    assert obj._idle_timeout == 30
    assert obj.server is not None
    last_timeout = FakeServer._last_timeout
    assert last_timeout >= 30
    failure_result = Target.RESULT_FAILURE
    obj._fixed_timeout = True
    (tmp_path / "lithium1").mkdir()
    assert obj.lithium_interesting(tmp_path / "lithium1")
    assert obj._idle_timeout == 30
    assert FakeServer._last_timeout == last_timeout
    obj._fixed_timeout = False
    (tmp_path / "lithium2").mkdir()
    assert obj.lithium_interesting(tmp_path / "lithium2")
    assert obj._idle_timeout < 30
    idle_timeout = obj._idle_timeout
    assert FakeServer._last_timeout == last_timeout
    last_timeout = FakeServer._last_timeout
    (tmp_path / "lithium3").mkdir()
    assert obj.lithium_interesting(tmp_path / "lithium3")
    assert obj._idle_timeout == idle_timeout
    assert FakeServer._last_timeout  # assert that there is actually a timeout
    assert FakeServer._last_timeout < last_timeout
    last_timeout = FakeServer._last_timeout
    assert obj.server is not None
    (tmp_path / "lithium4").mkdir()
    assert obj.lithium_interesting(tmp_path / "lithium4")
    assert obj._idle_timeout == idle_timeout
    assert FakeServer._last_timeout == last_timeout
    assert obj.server is not None
    obj.lithium_cleanup()


def test_idle_timeout(mocker, tmp_path):
    "idle is not polled until idle_timeout has passed, and idle_poll results in timeout"
    fake_ichk = mocker.patch("grizzly.common.runner._IdleChecker", autospec=True)
    obj = ReductionJob([], FakeTarget(), 30, False, False, 0, 1, 1, 25, 30, FakeReduceStatus(),
                       testcase_cache=False)
    create_target_binary(obj.target, tmp_path)
    (tmp_path / "test.html").touch()
    obj.reduce_file = str(tmp_path / "test.html")
    obj.lithium_init()
    (tmp_path / "lithium0").mkdir()
    assert obj.lithium_interesting(tmp_path / "lithium0")
    assert fake_ichk.called
    assert fake_ichk.call_args[0][1] == 25  # idle_threshold
    assert fake_ichk.call_args[0][2] == 30  # idle_timeout
    obj.lithium_cleanup()
